/*
 * Copyright © 2014 Red Hat, Inc
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *       Alexander Larsson <alexl@redhat.com>
 */

#include "config.h"

#include <locale.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "libgsystem.h"
#include "libglnx/libglnx.h"

#include "xdg-app-builtins.h"
#include "xdg-app-utils.h"

static char *opt_arch;

static GOptionEntry options[] = {
  { "arch", 0, 0, G_OPTION_ARG_STRING, &opt_arch, "Arch to install for", "ARCH" },
  { NULL }
};

gboolean
xdg_app_builtin_install_runtime (int argc, char **argv, GCancellable *cancellable, GError **error)
{
  gboolean ret = FALSE;
  g_autoptr(GOptionContext) context = NULL;
  g_autoptr(XdgAppDir) dir = NULL;
  g_autoptr(GFile) deploy_base = NULL;
  g_autoptr(GFile) origin = NULL;
  const char *repository;
  const char *runtime;
  const char *branch = "master";
  g_autofree char *ref = NULL;
  gboolean created_deploy_base = FALSE;

  context = g_option_context_new ("REPOSITORY RUNTIME [BRANCH] - Install a runtime");

  if (!xdg_app_option_context_parse (context, options, &argc, &argv, 0, &dir, cancellable, error))
    goto out;

  if (argc < 3)
    {
      usage_error (context, "REPOSITORY and RUNTIME must be specified", error);
      goto out;
    }

  repository = argv[1];
  runtime  = argv[2];
  if (argc >= 4)
    branch = argv[3];

  if (!xdg_app_is_valid_name (runtime))
    {
      xdg_app_fail (error, "'%s' is not a valid runtime name", runtime);
      goto out;
    }

  if (!xdg_app_is_valid_branch (branch))
    {
      xdg_app_fail (error, "'%s' is not a valid branch name", branch);
      goto out;
    }

  ref = xdg_app_build_runtime_ref (runtime, branch, opt_arch);

  deploy_base = xdg_app_dir_get_deploy_dir (dir, ref);
  if (g_file_query_exists (deploy_base, cancellable))
    {
      xdg_app_fail (error, "Runtime %s branch %s already installed", runtime, branch);
      goto out;
    }

  if (!xdg_app_dir_pull (dir, repository, ref,
                         cancellable, error))
    goto out;

  if (!g_file_make_directory_with_parents (deploy_base, cancellable, error))
    goto out;
  created_deploy_base = TRUE;

  origin = g_file_get_child (deploy_base, "origin");
  if (!g_file_replace_contents (origin, repository, strlen (repository), NULL, FALSE,
                                G_FILE_CREATE_NONE, NULL, cancellable, error))
    goto out;

  if (!xdg_app_dir_deploy (dir, ref, NULL, cancellable, error))
    goto out;

  xdg_app_dir_cleanup_removed (dir, cancellable, NULL);

  ret = TRUE;

 out:
  if (created_deploy_base && !ret)
    gs_shutil_rm_rf (deploy_base, cancellable, NULL);

  return ret;
}

gboolean
xdg_app_builtin_install_app (int argc, char **argv, GCancellable *cancellable, GError **error)
{
  gboolean ret = FALSE;
  g_autoptr(GOptionContext) context = NULL;
  g_autoptr(XdgAppDir) dir = NULL;
  g_autoptr(GFile) deploy_base = NULL;
  g_autoptr(GFile) origin = NULL;
  const char *repository;
  const char *app;
  const char *branch = "master";
  g_autofree char *ref = NULL;
  gboolean created_deploy_base = FALSE;

  context = g_option_context_new ("REPOSITORY APP [BRANCH] - Install an application");

  if (!xdg_app_option_context_parse (context, options, &argc, &argv, 0, &dir, cancellable, error))
    goto out;

  if (argc < 3)
    {
      usage_error (context, "REPOSITORY and APP must be specified", error);
      goto out;
    }

  repository = argv[1];
  app  = argv[2];
  if (argc >= 4)
    branch = argv[3];

  if (!xdg_app_is_valid_name (app))
    {
      xdg_app_fail (error, "'%s' is not a valid application name", app);
      goto out;
    }

  if (!xdg_app_is_valid_branch (branch))
    {
      xdg_app_fail (error, "'%s' is not a valid branch name", branch);
      goto out;
    }

  ref = xdg_app_build_app_ref (app, branch, opt_arch);

  deploy_base = xdg_app_dir_get_deploy_dir (dir, ref);
  if (g_file_query_exists (deploy_base, cancellable))
    {
      xdg_app_fail (error, "App %s branch %s already installed", app, branch);
      goto out;
    }

  if (!xdg_app_dir_pull (dir, repository, ref,
                         cancellable, error))
    goto out;

  if (!g_file_make_directory_with_parents (deploy_base, cancellable, error))
    goto out;
  created_deploy_base = TRUE;

  origin = g_file_get_child (deploy_base, "origin");
  if (!g_file_replace_contents (origin, repository, strlen (repository), NULL, FALSE,
                                G_FILE_CREATE_NONE, NULL, cancellable, error))
    goto out;

  if (!xdg_app_dir_deploy (dir, ref, NULL, cancellable, error))
    goto out;

  if (!xdg_app_dir_make_current_ref (dir, ref, cancellable, error))
    goto out;

  if (!xdg_app_dir_update_exports (dir, app, cancellable, error))
    goto out;

  xdg_app_dir_cleanup_removed (dir, cancellable, NULL);

  ret = TRUE;

 out:
  if (created_deploy_base && !ret)
    gs_shutil_rm_rf (deploy_base, cancellable, NULL);

  return ret;
}
